import statsmodels.api as sm
import pandas as pd
import numpy as np
from numpy import nan

from .pdutils import split

def reg_by(df, yvar, xvars, groupvar, merge=False, cons=True):
    """
    Runs a regression of df[yvar] on df[xvars] by values of groupvar. Outputs a dataframe with values of
    groupvar and corresponding coefficients, unless merge=True, then outputs the original dataframe with the
    appropriate coefficients merged in.

    Required inputs:
    groupvar: str or list of strs, column names of columns identifying by groups

    Optional Options:
    cons: True to include a constant, False to not
    """
    # Convert xvars to list if str is passed
    xvars = _check_inputs_regby(xvars)

    # If there are multiple groupvars, create a key variable containing all groupvars (modifies df inplace)
    groupvar, drop_group = _set_groupvar_and_drop_group(df, groupvar)

    # Select dataframe of only y, x and group vars and drop any missings
    yx_df = df_for_reg(df, yvar, xvars, groupvar)

    # Create a list of right hand side variables. Includes 'const' if cons is True
    rhs = _set_rhs(xvars, cons)

    # Split DataFrame into a list of arrays with each bygroup being one array. Provide an accompanying list of bygroups
    arrs, groups = _get_lists_of_arrays_and_groups(yx_df, yvar, xvars, groupvar)

    # Run regressions by groups, storing results as a list of numpy arrays
    results = _reg_by(arrs, groups, xvars, rhs, cons)

    # Combine list of arrays into df, and apply column labels
    result_df = _result_list_of_arrays_to_df(results, rhs, groupvar)

    if merge:
        result_df = df.merge(result_df, how='left', on=groupvar)
    if drop_group:
        result_df.drop(groupvar, axis=1, inplace=True)

    return result_df.reset_index(drop=True)

def df_for_reg(df, yvar, xvars, groupvar):
    # Select dataframe of only y and x vars
    yx_df = df.loc[:, xvars + [yvar]]
    # Recombine groupvar and drop missing
    yx_df = pd.concat([yx_df, df[groupvar]], axis=1).dropna()

    return yx_df

def _reg_by(arrs, groups, xvars, rhs, cons):
    results = []
    for i, arr in enumerate(arrs):
        results.append(_reg(arr, xvars, rhs, cons, groups[i]))
    return results

def _reg(arr, xvars, rhs, cons, group):
    X = arr[:, 1:].astype(float)

    if cons:
        X = sm.add_constant(X)

    y = arr[:, 0].astype(float)

    if arr.shape[0] > len(xvars) + 1:  # if enough observations, run regression
        model = sm.OLS(y, X)
        result = model.fit()
        this_result = np.append(result.params, group)  # add groupvar
        this_result = this_result[None, :]  # cast 1d array into 2d array
    else:  # not enough obs, return nans
        this_result = np.empty((1, len(rhs) + 1), dtype='O')
        this_result[:] = nan
        this_result[0, len(rhs)] = group

    return this_result


def _check_inputs_regby(xvars):
    if isinstance(xvars, str):
        xvars = [xvars]
    assert isinstance(xvars, list)

    return xvars

def _set_groupvar_and_drop_group(df, groupvar):
    drop_group = False
    if isinstance(groupvar, list):
        df['__key_regby__'] = ''
        for var in groupvar:
            df['__key_regby__'] = df['__key_regby__'] + df[var].astype(str)
        groupvar = '__key_regby__'
        drop_group = True

    return groupvar, drop_group

def _set_rhs(xvars, cons):
    if cons:
        rhs = ['const'] + xvars
    else:
        rhs = xvars

    return rhs

def _result_list_of_arrays_to_df(results, rhs, groupvar):
    result_df = pd.DataFrame(np.concatenate(results, axis=0))
    result_df = result_df.apply(pd.to_numeric, errors='ignore')
    cols = rhs + [groupvar]
    result_df.columns = ['coef_' + col if col not in (groupvar, 'const') else col for col in cols]

    return result_df

def _get_lists_of_arrays_and_groups(df, yvar, xvars, groupvar):
    arrs = split(df, [yvar] + xvars, keyvar=groupvar)
    groups = df[groupvar].unique().tolist()
    assert len(arrs) == len(groups)

    return arrs, groups





def _to_list_if_str(list_):
    return _to_list_if_type(list_, str)

def _to_list_if_tuple(list_):
    return _to_list_if_type(list_, tuple)

def _to_list_if_type(list_, type):
    if isinstance(list_, type):
        list_ = [list_]
    assert isinstance(list_, list)
    return list_



def extract_all_dummy_cols_from_dummy_cols_dict(dummy_cols_dict):
    all_values = []
    for fe_col, fe_values in dummy_cols_dict.items():
        all_values += fe_values
    return all_values

def extract_all_dummy_cols_from_dummy_cols_dict_list(dummy_cols_dict_list):
    all_cols = []
    for dummy_cols_dict in dummy_cols_dict_list:
        if dummy_cols_dict:
            all_cols += extract_all_dummy_cols_from_dummy_cols_dict(dummy_cols_dict)
    return list(set(all_cols))


def extract_all_fe_names_from_dummy_cols_dict_list(dummy_cols_dict_list):
    names = []
    for dummy_cols_dict in dummy_cols_dict_list:
        if dummy_cols_dict:
            names += [name for name in dummy_cols_dict]

    return list(set(names))


import warnings

from .reg import reg
from .differenced import diff_reg


def any_reg(reg_type, *reg_args, **reg_kwargs):

    reg_args, reg_kwargs = _validate_inputs(*reg_args, **reg_kwargs)

    reg_type = reg_type.lower()

    if _is_diff_reg_str(reg_type):
        temp_kwargs = reg_kwargs.copy()
        if 'diff_cols' in temp_kwargs:
            diff_cols = temp_kwargs.pop('diff_cols')
        else:
            diff_cols = None
        id_col = temp_kwargs.pop('id_col')
        date_col = temp_kwargs.pop('date_col')
        return diff_reg(*reg_args, id_col, date_col, diff_cols=diff_cols, **temp_kwargs)

    if _is_normal_reg_str(reg_type):
        return reg(*reg_args, **reg_kwargs)

    raise ValueError(f'Must pass valid reg type. Got {reg_type}')

def _is_diff_reg_str(reg_str):
    return reg_str in ('diff', 'difference', 'diff_reg', 'diff reg', 'difference reg', 'difference regression')

def _is_normal_reg_str(reg_str):
    return reg_str in ('reg', 'normal', 'ols') or reg_str == None

def _validate_inputs(*args, **kwargs):
    yvar = args[1]
    xvars = args[2]

    if yvar in xvars:
        warnings.warn(f'{yvar} is both Y variable and passed in X variables. Removing from X for this model.', UserWarning)
        new_xvars = xvars.copy()
        new_xvars.remove(yvar)
        args = args[:2] + (new_xvars,) + args[3:]

    return args, kwargs



import pandas as pd


def add_fake_byvars(df, data_multipler_per_byvar=2, num_byvars=1, byvar_basename='byvar'):
    """
    WARNING: do not use when index is important. index will be dropped.

    Duplicates data and creates a new variables in each copy of the data. Then combines the two
    so that the new dataset is twice the size, with repeated data except for one new variable
    which just says which duplicated dataset it came from.

    Then if num_byvars > 1, does the above repeatedly, creating new byvars. The total data
    size will be n * data_multipler_per_byvar ^ num_byvars

    Useful for performance testing when a function does something in groups by a variable.
    """

    for i in range(num_byvars):
        df = _add_fake_byvar(df, data_multipler=data_multipler_per_byvar, byvar_name=byvar_basename)

    return df

def _add_fake_byvar(df, data_multipler=2, byvar_name='byvar'):
    """
    WARNING: do not use when index is important. index will be dropped.

    Duplicates data and creates a new variable in each copy of the data. Then combines the two
    so that the new dataset is twice the size, with repeated data except for one new variable
    which just says which duplicated dataset it came from.

    Useful for performance testing when a function does something in groups by a variable.
    """
    df_dups = [df.copy() for i in range(data_multipler)]
    byvar_name = _set_byvar_name(df.columns, byvar_name=byvar_name)

    out_df = pd.DataFrame()
    for i, temp_df in enumerate(df_dups):
        temp_df[byvar_name] = i
        out_df = out_df.append(temp_df)

    return out_df.reset_index(drop=True)


def _set_byvar_name(cols, byvar_name='byvar', i=0):

    if i == 0:
        # try original before adding index
        new_byvar_name = byvar_name
    else:
        new_byvar_name = f'{byvar_name}{i}'

    if new_byvar_name in cols:
        i += 1
        return _set_byvar_name(cols, byvar_name=byvar_name, i=i)
    else:
        return new_byvar_name


import pandas as pd


from ..tools import _to_list_if_str

def fixed_effects_reg_df_and_cols_dict(df, fe_vars):
    fe_vars = _to_list_if_str(fe_vars)

    fe_cols_dict = {}
    for fe_var in fe_vars:
        df, cols = _fixed_effects_reg_df_and_cols(df, fe_var)
        fe_cols_dict[fe_var] = cols

    return df, fe_cols_dict

def _fixed_effects_reg_df_and_cols(df, fe_var):
    dummies = pd.get_dummies(df[fe_var].astype(str)).iloc[:,1:] #drop first dummy, this will be the excluded group
    dummy_cols = [col for col in dummies.columns]
    return pd.concat([df, dummies], axis=1).drop(fe_var, axis=1), dummy_cols




import itertools

from .chooser import any_reg, _is_diff_reg_str
from .summarize import produce_summary
from .select import select_models


def reg_for_each_combo(df, yvar, xvars, reg_type='reg', **reg_kwargs):
    """
    Takes each possible combination of xvars (starting from each var individually, then each pair
    of vars, etc. all the way up to all xvars), and regresses yvar on each set of xvars. Returns
    a list of fitted regressions.
    """
    reg_list = []
    for i in range(1, len(xvars) + 1):
        for combo in itertools.combinations(xvars, i):
            x = list(combo)
            reg_list.append(any_reg(reg_type, df, yvar, x, **reg_kwargs))

    return reg_list


def reg_for_each_xvar_set(df, yvar, xvars_list, reg_type='reg', **reg_kwargs):
    """
    Runs regressions on the same y variable for each set of x variables passed. xvars_list
    should be a list of lists, where each individual list is one set of x variables for one model.
    Returns a list of fitted regressions.

    If fe is passed, should either pass a string to use fe in all models, or a list of strings or
    None of same length as num models
    """
    fe, interaction_tuples = _pop_and_convert_kwargs_which_are_repeated_across_models(reg_kwargs, len(xvars_list))
    return [any_reg(reg_type, df, yvar, x, fe=fe[i], interaction_tuples=interaction_tuples[i], **reg_kwargs) for i, x in enumerate(xvars_list)]


def reg_for_each_xvar_set_and_produce_summary(df, yvar, xvars_list, robust=True,
                                              cluster=False, stderr=False, fe=None, float_format='%0.2f',
                                              suppress_other_regressors=False,
                                              regressor_order=[], **other_reg_kwargs):
    """
    Convenience function to run regressions for every set of xvars passed
    and present them in a summary format. Returns a tuple of (reg_list, summary) where reg_list
    is a list of fitted regression models, and summary is a single dataframe of results.

    Required inputs:
    df: pandas dataframe containing regression data
    yvar: str, column name of y variable
    xvars_list: list of lists of strs, each individual list has column names of x variables for that model

    Optional inputs:
    robust: bool, set to True to use heterskedasticity-robust standard errors
    cluster: False or str, set to a column name to calculate standard errors within clusters
             given by unique values of given column name
    stderr: bool, set to True to keep rows for standard errors below coefficient estimates
    fe:    If fe is passed, should either pass a string to use fe in all models, or a list of strings or
    None of same length as num models

    Note: only specify at most one of robust and cluster.
    """
    reg_list = reg_for_each_xvar_set(df, yvar, xvars_list, robust=robust, cluster=cluster, fe=fe, **other_reg_kwargs)
    regressor_order = _set_regressor_order(regressor_order, other_reg_kwargs)
    summ = produce_summary(reg_list, stderr=stderr, float_format=float_format,
                           regressor_order=regressor_order, suppress_other_regressors=suppress_other_regressors)
    return reg_list, summ

def reg_for_each_yvar_and_produce_summary(df, yvars, xvars, main_ivs, reg_type='reg',
                                          stderr=False, float_format='%0.2f',
                                          **reg_kwargs):
    reg_list = reg_for_each_yvar(df, yvars, xvars, reg_type=reg_type, **reg_kwargs)
    regressor_order = _set_regressor_order(main_ivs, reg_kwargs)
    summ = produce_summary(reg_list, stderr=stderr, float_format=float_format,
                           regressor_order=regressor_order, suppress_other_regressors=True)

    # Transposed is a better layout since all controls are suppressed and there may be many models
    summ.tables[0] = summ.tables[0].T

    return reg_list, summ



def reg_for_each_yvar(df, yvars, xvars, reg_type='reg', **reg_kwargs):
    return [any_reg(reg_type, df, yvar, xvars, **reg_kwargs) for yvar in yvars]


def reg_for_each_combo_select_and_produce_summary(df, yvar, xvars, robust=True, cluster=False,
                                                  keepnum=5, stderr=False, float_format='%0.1f',
                                                  regressor_order=[],
                                                  **other_reg_kwargs):
    """
    Convenience function to run regressions for every combination of xvars, select the best models,
    and present them in a summary format. Returns a tuple of (reg_list, summary) where reg_list
    is a list of fitted regression models, and summary is a single dataframe of results

    Required inputs:
    df: pandas dataframe containing regression data
    yvar: str, column name of y variable
    xvars: list of strs, column names of all possible x variables

    Optional inputs:
    robust: bool, set to True to use heterskedasticity-robust standard errors
    cluster: False or str, set to a column name to calculate standard errors within clusters
             given by unique values of given column name
    keepnum: int, number to keep for each amount of x variables. The total number of outputted
             regressions will be roughly keepnum * len(xvars)
    stderr: bool, set to True to keep rows for standard errors below coefficient estimates

    Note: only specify at most one of robust and cluster.

    """
    reg_list = reg_for_each_combo(df, yvar, xvars, robust=robust, cluster=cluster, **other_reg_kwargs)
    regressor_order = _set_regressor_order(regressor_order, other_reg_kwargs)
    outlist = select_models(reg_list, keepnum, xvars)
    summ = produce_summary(outlist, stderr=stderr, float_format=float_format, regressor_order=regressor_order)
    return outlist, summ

def _pop_and_convert_kwargs_which_are_repeated_across_models(reg_kwargs, num_models):
    if 'fe' in reg_kwargs:
        fe = reg_kwargs.pop('fe')
    else:
        fe = None

    if 'interaction_tuples' in reg_kwargs:
        interaction_tuples = reg_kwargs.pop('interaction_tuples')
    else:
        interaction_tuples = []

    fe = _set_fe(fe, num_models)
    interaction_tuples = _set_interaction_tuples(interaction_tuples, num_models)

    return fe, interaction_tuples


def _set_fe(fe, num_models):
    return _set_for_multiple_models(fe, num_models, param_name='fixed effects')

def _set_interaction_tuples(interaction_tuples, num_models):
    return _set_for_multiple_models(interaction_tuples, num_models, param_name='interaction tuples')

def _set_for_multiple_models(param, num_models, param_name='fixed effects'):
    # Here we are being passed a list of strings or None matching the size of models.
    # This is the correct format so just output
    if (isinstance(param, list)) and (len(param) == num_models):
        out_param = param

    # Here we are being passed a single item or a list with a single item
    # Need to expand to cover all models
    else:
        if (not isinstance(param, list)):
            param = [param]
        if len(param) > 1:
            raise ValueError(
                f'Incorrect shape of items for {param_name} passed. Got {len(param)} items, was expecting {num_models}')
        out_param = [param[0]] * num_models

    # Final input checks
    assert isinstance(out_param, list)

    return out_param

def _set_regressor_order(regressor_order, reg_kwargs):
    # No processing needed if not difference regression
    if ('reg_type' not in reg_kwargs) or (not _is_diff_reg_str(reg_kwargs['reg_type'])):
        return regressor_order

    if 'diff_cols' in reg_kwargs:
        cols = reg_kwargs['diff_cols']
    else:
        cols = 'all'

    return _convert_regressor_order_for_diff(regressor_order, cols)


def _convert_regressor_order_for_diff(regressor_order, cols='all'):
    if cols == 'all':
        cols = regressor_order.copy()

    return [col + ' Change' if col in cols else col for col in regressor_order]


import pandas as pd
import numpy as np
from functools import partial
from itertools import product

from .pdutils import _to_list_if_str

def fillna_by_groups_and_keep_one_per_group(df, byvars, exclude_cols=None, str_vars='first', num_vars='mean'):
    """
    WARNING: do not use if index is important, it will be dropped
    """
    byvars = _to_list_if_str(byvars)
    if exclude_cols:
        exclude_cols = _to_list_if_str(exclude_cols)

    df = fillna_by_groups(df, byvars, exclude_cols=exclude_cols, str_vars=str_vars, num_vars=num_vars)
    _drop_duplicates(df, byvars)

    return df


def fillna_by_groups(df, byvars, exclude_cols=None, str_vars='first', num_vars='mean'):
    """
    WARNING: do not use if index is important, it will be dropped
    """
    byvars = _to_list_if_str(byvars)

    if exclude_cols:
        cols_to_fill = [col for col in df.columns if (col not in exclude_cols) and (col not in byvars)]
        concat_vars = byvars + exclude_cols
    else:
        cols_to_fill = [col for col in df.columns if col not in byvars]
        concat_vars = byvars

    _fill_data = partial(_fill_data_for_series, str_vars=str_vars, num_vars=num_vars)

    out_dfs = []
    for group, group_df in df[byvars + cols_to_fill].groupby(byvars):
        out_dfs.append(group_df.reset_index(drop=True).transform(_fill_data))

    filled = pd.concat(out_dfs, axis=0).reset_index(drop=True)

    filled = _restore_nans_after_fill(filled) #_fill_data places -999.999 in place of nans, now convert back


    return filled

def add_missing_group_rows(df, fill_id_cols):
    fill_ids = [df[fill_id_col].unique() for fill_id_col in fill_id_cols]
    index_df = pd.DataFrame([i for i in product(*fill_ids)], columns=fill_id_cols)

    return index_df.merge(df, how='left', on=fill_id_cols)

def drop_missing_group_rows(df, fill_id_cols):
    drop_subset = [col for col in df.columns if col not in fill_id_cols]
    return df.dropna(subset=drop_subset, how='all')

def _fill_data_for_series(series, str_vars='first', num_vars='mean'):
    index = _get_non_nan_value_index(series, str_vars)
    if index is None:
        # All nans, can't do anything but return back nothing
        # But transform ignores nans in the output and then complains when the sizes don't match.
        # So instead, put a placeholder of -999.999
        return pd.Series([-999.999 for i in range(len(series))])
    # handle numeric
    if series.dtype in (np.float64, np.int64):
        if num_vars in ('first', 'last'):
            # Overwrite index for that of num vars if not using the same value as for str vars
            if num_vars != str_vars:
                index = _get_non_nan_value_index(series, num_vars)
            return _fill_data_for_str_series(series, non_nan_index=index)
        return _fill_data_for_numeric_series(series, fill_function=num_vars)
    # handle strs
    else:
        return _fill_data_for_str_series(series, non_nan_index=index)


def _fill_data_for_numeric_series(series, fill_function='mean'):
    return series.fillna(series.agg(fill_function))


def _fill_data_for_str_series(series, non_nan_index):
    fill_value = series.loc[non_nan_index]

    return series.fillna(fill_value)

def _get_non_nan_value_index(series, first_or_last):
    if first_or_last == 'first':
        return series.first_valid_index()
    elif first_or_last == 'last':
        return series.last_valid_index()
    else:
        raise ValueError("Did not pass 'first' or 'last'")


def _restore_nans_after_fill(df):
    """
    -999.999 was used as a missing representation as pandas can not handle nans in transform.
    Convert back to nan now
    """
    return df.applymap(lambda x: np.nan if x == -999.999 else x)

def _drop_duplicates(df, byvars):
    """
    Note: inplace
    """
    df.drop_duplicates(subset=byvars, inplace=True)